{-------------------------------------------------------------------------------
  Visit comment structure
  The core of the builder. It takes data from the parser, and build comments,
  and TAG comments

  http://dephicodetodoc.sourceforge.net/
  Copyright TridenT 2003 - Under GNU GPL licence

  @Author    TridenT
  @Version   2003/12/01   TridenT   v0.0     Initial revision
-------------------------------------------------------------------------------}


unit VisitComment;

interface

uses BaseVisitor, VisitParseTree, Structure, classes, ParseTreeNode, ParseTreeNodeType, Tokens;


type
  TVInterfaceComment = class(TBaseTreeNodeVisitor)
  private
    mDebug: string;
    procedure GetTAG(const tagNode: TObject);
    function GetUnitTemplate(const TaggedNode: TParseTreeNode): TDOCTemplate;
    function GetProcedureTemplate(const TaggedNode: TParseTreeNode): TDOCTemplate;
    function GetIdentifierTemplate(const TaggedNode: TParseTreeNode): TDOCTemplate;
    procedure ExtractTAG(const etNode: TObject; const DocTemplate: TDOCTemplate);
    procedure GetClassType(const ctNode: TObject);
    procedure GetInterfaceType(const ctNode: TObject);
    procedure GetVarDecl(const ctNode: TObject);
  protected
  public
    constructor Create; override;
    procedure PreVisitParseTreeNode(const pcNode: TObject;
      var prVisitResult: TRVisitResult); override;
    procedure PostVisitParseTreeNode(const pcNode: TObject); override;
    procedure VisitSourceToken(const pcNode: TObject; var prVisitResult: TRVisitResult);
      override;
    function FinalSummary(var psMessage: string): boolean; override;
    function IsIncludedInSettings: boolean; override;
  end;

implementation

uses
  { delphi }
  SysUtils,
  JclStrings,
  { JCF  }
  SourceToken, TokenUtils, uDocProject, TAGComment;

var
  CurrentUnit: TDOCUnit = nil;
  CurrentClass: TDOCClass = nil;
  CurrentMethod: TDOCMethod = nil;
  CurrentFunction: TDOCFunction = nil;
  LastTaggedNode: TParseTreeNode = nil;

  { TVInterfaceStructure }

function TVInterfaceComment.FinalSummary(var psMessage: string): boolean;
begin
  //
  psMessage := mDebug;
  Result    := True;
end;

function TVInterfaceComment.IsIncludedInSettings: boolean;
begin
  Result := True;
end;

procedure TVInterfaceComment.PreVisitParseTreeNode(const pcNode: TObject;
  var prVisitResult: TRVisitResult);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(pcNode);
  // For the moment, only Interface section is analysed, skip Implementation
  //if lcNode.HasParentNode(nImplementationSection) then exit;
  if lcNode.NodeType = nUnitName then
    CurrentUnit := DocProject.GetStructure.FindUnit(lcNode.FirstSolidLeaf.Describe)
        as TDOCUnit;
  if lcNode.NodeType = nClassType then GetClassType(lcNode);
  if lcNode.NodeType = nInterfaceType then GetInterfaceType(lcNode);
  if lcNode.NodeType = nVarDecl then GetVarDecl(lcNode);

  if lcNode.NodeType in TAGNodes then GetTAG(lcNode);
end;

procedure TVInterfaceComment.GetTAG(const tagNode: TObject);
var
  lcNode: TParseTreeNode;
  TaggedNode:  TParseTreeNode;
  DocTemplate : TDOCTemplate;
begin
  lcNode := TParseTreeNode(tagNode);
  Assert(lcNode.HasParentNode([nCurlyBlock,nInlineComment,nEndOfLineComment]));

  // Get CurlyBlock or Inline parent's Node
  // Warning : Inline can exist AFTER the parent (same line)
  if lcNode.HasParentNode([nCurlyBlock,nInlineComment]) then
   TaggedNode:=lcNode.GetParentNode([nCurlyBlock,nInlineComment]).Parent
  else
   if lcNode.HasParentNode([nEndOfLineComment]) then
    TaggedNode:=LastTaggedNode
   else TaggedNode:=nil;

  assert(Assigned(TaggedNode));
  //Options for Curly Block only
  if lcNode.HasParentNode(nCurlyBlock) then
  begin
    // Verify if it is First or last TAG comment of the Curly block
    assert(lcNode.PriorLeafNode is TSourceToken);
    assert(lcNode.NextLeafNode is TSourceToken);
    // Verify if FirstTAG must be skipped
    if DocProject.DocOptions.SkipFirstTAGLine then
      if TSourceToken(lcNode.PriorLeafNode).SourceCode = '{' then
       exit;
    // Verify if LastTAG must be skip
    if DocProject.DocOptions.SkipLastTAGLine then
      if TSourceToken(lcNode.NextLeafNode).SourceCode = '}' then
       exit;
  end;

  case TaggedNode.NodeType of
   nUnitHeader: DOCTemplate:=GetUnitTemplate(TaggedNode);
   nFunctionHeading: DOCTemplate:=GetProcedureTemplate(TaggedNode);
   nConstructorHeading: DOCTemplate:=GetProcedureTemplate(TaggedNode);
   nDestructorHeading: DOCTemplate:=GetProcedureTemplate(TaggedNode);
   nProcedureHeading: DOCTemplate:=GetProcedureTemplate(TaggedNode);
   nIdentifier: DOCTemplate:=GetIdentifierTemplate(TaggedNode);

   else DocTemplate:=nil;
  end;

  if assigned(DocTemplate) then
  begin
    ExtractTAG(tagNode, DocTemplate);
  end;

end;

// Source Tokens are only LEAF !
procedure TVInterfaceComment.VisitSourceToken(const pcNode: TObject;
  var prVisitResult: TRVisitResult);
begin
end;

constructor TVInterfaceComment.Create;
begin
  mDebug := '';
end;

procedure TVInterfaceComment.PostVisitParseTreeNode(const pcNode: TObject);
begin
  //
end;

function TVInterfaceComment.GetUnitTemplate(const TaggedNode: TParseTreeNode): TDOCTemplate;
begin
  result:=DocProject.GetStructure.FindUnit(TaggedNode.GetImmediateChild(nUnitName).FirstSolidLeaf.Describe);
end;

procedure TVInterfaceComment.ExtractTAG(const etNode: TObject; const DocTemplate: TDOCTemplate);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(etNode);
  DocTemplate.TAGHolderList.AddTAG(TSourceToken(lcNode.FirstLeaf).SourceCode
   ,lcNode.NodeType);
end;

function TVInterfaceComment.GetProcedureTemplate(
  const TaggedNode: TParseTreeNode): TDOCTemplate;
var
  IdentifierNode, ClassNode : TParseTreeNode;
begin
  Result:=nil;
  if TaggedNode.HasParentNode(nInterfaceSection) then exit;
  // Get Unit
  assert(CurrentUnit<>nil);
  // Get ClassName if exist

  // Get Identifier
  Assert(TaggedNode.HasChildNode(nIdentifier));
  // WARNING! Simple function (not member) have no identifier node!
  // Corrected in TBuildParseTree.RecogniseMethodName by adding a Identifier node

  IdentifierNode:=TaggedNode.GetImmediateChild(nIdentifier).FirstSolidLeaf;
  if TSourceToken(IdentifierNode.NextLeafNode).SourceCode='.' then
  begin
    ClassNode:= IdentifierNode;
    // Found the Class name of a method
    IdentifierNode:=ClassNode.NextLeafNode.NextLeafNode;
    CurrentClass:= CurrentUnit.FindClass(TSourceToken(ClassNode).SourceCode) as TDOCClass;
    Result:=CurrentClass.FindMember(TSourceToken(IdentifierNode).SourceCode);
  end
  else // so it's a simple function, not method
  result:=nil;
end;

function TVInterfaceComment.GetIdentifierTemplate(
  const TaggedNode: TParseTreeNode): TDOCTemplate;
var
  IdentifierNode : TParseTreeNode;
begin
  Result:=nil;
  IdentifierNode:= TaggedNode.FirstSolidLeaf;
  // Assume it can only be a CLASS definition
  // But it can be another type !

  if TaggedNode.HasParentNode(nClassDeclarations) then
    Result:= CurrentClass.FindMember(TSourceToken(IdentifierNode).SourceCode)
  else
    if TaggedNode.HasParentNode(nTypeDecl) then
      Result:= CurrentUnit.FindClass(TSourceToken(IdentifierNode).SourceCode);
end;

procedure TVInterfaceComment.GetClassType(const ctNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ctNode);
  // Verify if it is a forward declaration
  if (lcNode.HasChildNode(nClassBody) or lcNode.HasChildNode(nClassHeritage)) then
  begin
    Assert(CurrentUnit <> nil);
    CurrentClass := CurrentUnit.FindClass(lcNode.GetParentNode(nTypeDecl).GetImmediateChild
      (nIdentifier).FirstSolidLeaf.Describe) as TDOCClass;
  end;
end;

procedure TVInterfaceComment.GetInterfaceType(const ctNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ctNode);
  // Verify if it is a forward declaration
  if (lcNode.HasChildNode(nInterfaceBody) or lcNode.HasChildNode(nInterfaceHeritage)) then
  begin
    Assert(CurrentUnit <> nil);
    CurrentClass := CurrentUnit.FindClass(lcNode.GetParentNode(nTypeDecl).GetImmediateChild
      (nIdentifier).FirstSolidLeaf.Describe) as TDOCClass;
  end;
end;

procedure TVInterfaceComment.GetVarDecl(const ctNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ctNode);
  // Get First Identifier
  if lcNode.HasChildNode(nIdentifier) then
   LastTaggedNode:=lcNode.FirstSolidLeaf.GetParentNode(nIdentifier);

end;

end.
