{-------------------------------------------------------------------------------
  Visit inteface structure
  The core of the builder. It takes data from the parser, and build class,
  units, and members hierarchy.
  http://dephicodetodoc.sourceforge.net/
  Copyright TridenT 2003 - Under GNU GPL licence

  @Author    TridenT
  @Version   2003/12/01   TridenT   Initial revision
-------------------------------------------------------------------------------}
unit VisitStructure;

interface

uses BaseVisitor, VisitParseTree, Structure, classes;

type
  // Visitor that build the structure from the result of the parser
  TVInterfaceStructure = class(TBaseTreeNodeVisitor)
  private
    mDebug: string; // Debug message writtent at the end of the process
    procedure GetClassMethod(const mNode: TObject; const mt: TMethodType);
    function GetSourceCode(const tkNode: TObject): string;
    function GetIdentList(const idNode: TObject; const idList: TStringList): integer;
    procedure GetFormalParameter(const fpNode: TObject);
    procedure GetReturnType(const rtNode: TObject);
    procedure GetClassType(const ctNode: TObject);
    procedure GetInterfaceType(const ctNode: TObject);
    procedure GetClassHeritage(const ceNode: TObject);
    procedure GetClassProperty(const cpNode: TObject);
    procedure GetClassField(const cfNode: TObject);
    procedure GetUnitVar(const ivNode: TObject);
    procedure GetUnitConst(const icNode: TObject);
    procedure GetMemberVisibility(const mvNode: TObject; const DocMember: TDOCMember);
    procedure GetUnitType(const itNode: TObject);
  protected
  public
    constructor Create; override;
    procedure PreVisitParseTreeNode(const pcNode: TObject;
      var prVisitResult: TRVisitResult); override;
    procedure PostVisitParseTreeNode(const pcNode: TObject); override;
    procedure VisitSourceToken(const pcNode: TObject; var prVisitResult: TRVisitResult);
      override;
    function FinalSummary(var psMessage: string): boolean; override;
    function IsIncludedInSettings: boolean; override;
  end;

implementation

uses
  { delphi }
  SysUtils,
  JclStrings,
  { JCF  }
  SourceToken, Tokens, ParseTreeNode, ParseTreeNodeType, TokenUtils, uDocProject, uOptions;

var
  CurrentUnit: TDOCUnit = nil;
  CurrentClass: TDOCClass = nil;
  CurrentMethod: TDOCMethod = nil;
  CurrentFunction: TDOCFunction = nil;

  { TVInterfaceStructure }

function TVInterfaceStructure.FinalSummary(var psMessage: string): boolean;
begin
  //
  psMessage := mDebug;
  Result    := True;
end;

function TVInterfaceStructure.IsIncludedInSettings: boolean;
begin
  Result := True;
end;

procedure TVInterfaceStructure.PreVisitParseTreeNode(const pcNode: TObject;
  var prVisitResult: TRVisitResult);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(pcNode);
  // For the moment, only Interface section is analysed, skip Implementation
  if lcNode.HasParentNode(nImplementationSection) then exit;

  case lcNode.NodeType of
    nTypeDecl: GetUnitType(lcNode);
    nConstDecl: GetUnitConst(lcNode);
    nProcedureHeading: GetClassMethod(lcNode, mtProcedure);
    nFunctionHeading: GetClassMethod(lcNode, mtFunction);
    nConstructorHeading: GetClassMethod(lcNode, mtConstructor);
    nDestructorHeading: GetClassMethod(lcNode, mtDestructor);
    nProperty: GetClassProperty(lcNode);
    nVarDecl: if lcNode.HasParentNode(ObjectBodies) then GetClassField(lcNode)
      else if lcNode.HasParentNode(nVarSection) then GetUnitVar(lcNode);
      nClassHeritage: GetClassHeritage(lcNode);
    nClassType: GetClassType(lcNode);
    nInterfaceType: GetInterfaceType(lcNode);
    nUnitName: CurrentUnit := DocProject.GetStructure.AddUnit(lcNode.FirstSolidLeaf.Describe)
        as TDOCUnit;
    nFormalParam: GetFormalParameter(lcNode);
    nFunctionReturnType: GetReturnType(lcNode);
    else; // no nothing
  end;
end;

// Source Tokens are only LEAF !
procedure TVInterfaceStructure.VisitSourceToken(const pcNode: TObject;
  var prVisitResult: TRVisitResult);
begin
end;

constructor TVInterfaceStructure.Create;
begin
  mDebug := '';
end;

function TVInterfaceStructure.GetSourceCode(const tkNode: TObject): string;
var
  ptNode: TParseTreeNode;

  function RecurseGetSource(const stNode: TParseTreeNode): string;
  var
    liLoop: integer;
  begin
    Result := '';
    if stNode.IsLeaf then
    begin
      if TSourceToken(stNode).IsSolid then
        Result := TSourceToken(stNode).SourceCode + ' ';
    end;

    if stNode.HasChildren then
      for liLoop := 0 to stNode.ChildNodeCount - 1 do
        Result := Result + RecurseGetSource(stNode.ChildNodes[liLoop]);
  end;
begin
  ptNode := TParseTreeNode(tkNode);
  Result := RecurseGetSource(ptNode);
end;

procedure TVInterfaceStructure.PostVisitParseTreeNode(const pcNode: TObject);
begin
  //
end;

function TVInterfaceStructure.GetIdentList(const idNode: TObject;
  const idList: TStringList): integer;
var
  lcNode: TParseTreeNode;
  IndexChild: integer;
begin
  assert(idList <> nil);
  lcNode := TParseTreeNode(idNode);
  for IndexChild := 0 to lcNode.ChildNodeCount - 1 do
  begin
    if lcNode.ChildNodes[IndexChild].NodeType = nIdentifier then
      idList.Add(lcNode.ChildNodes[IndexChild].FirstSolidLeaf.Describe);
  end;
  Result := idList.Count;
end;

procedure TVInterfaceStructure.GetClassMethod(const mNode: TObject;
  const mt: TMethodType);
var
  lcNode: TParseTreeNode;
  tmpNode: TParseTreeNode;
  tmpMethod: TDOCMethod;
  tmpFunction: TDOCFunction;
begin
  lcNode := TParseTreeNode(mNode);
  if lcNode.HasParentNode(ObjectBodies) then
  // method
  begin
    Assert(CurrentClass <> nil);
    tmpNode := (lcNode.FirstSolidLeaf as TSOurceToken).NextSolidToken;
    Assert(tmpNode <> nil);
    tmpMethod := CurrentClass.AddMethod(tmpNode.Describe) as TDOCMethod;
    tmpMethod.MethodType := mt;
    // Get method Visibility
    GetMemberVisibility(lcNode, tmpMethod);
    tmpMethod.BodySource := GetSourceCode(lcNode);
    CurrentMethod        := tmpMethod;
  end
  else
  // normal procedure / function
  begin
    Assert(CurrentUnit <> nil);
    if lcNode.HasParentNode(nTypeDecl) then
      if lcNode.GetParentNode(nTypeDecl).HasChildNode(nIdentifier) then
      begin
        tmpNode     := lcNode.GetParentNode(nTypeDecl).GetImmediateChild(nIdentifier).FirstSolidLeaf;
        tmpFunction := CurrentUnit.AddFunction(tmpNode.Describe) as TDOCFunction;
        tmpFunction.BodySource := GetSOurceCode(lcNode);
        //tmpFunction.MethodType:=mt;
        CurrentFunction := tmpFunction;
      end;
  end;
end;

procedure TVInterfaceStructure.GetFormalParameter(const fpNode: TObject);
var
  lcNode: TParseTreeNode;
  tmpParam: TDOCParameter;
  tmpIdentList: TStringList;
  IndexIdent: integer;
begin
  lcNode := TParseTreeNode(fpNode);

  if (lcNode.HasParentNode(ObjectBodies)) then
  begin
    Assert(CurrentMethod <> nil);
    tmpIdentList := TStringList.Create;
    GetIdentList(lcNode.GetImmediateChild(nIdentList), tmpIdentList);
    for IndexIdent := 0 to tmpIdentList.Count - 1 do
    begin
      tmpParam := CurrentMethod.AddParameter(TmpIdentList.Strings[IndexIdent]) as
        TDOCParameter;
      tmpParam.BodySource := GetSOurceCode(lcNode);
      // Verify if Parameter have a type
      if lcNode.HasChildNode(nType) then
        tmpParam.RelatedType := lcNode.GetImmediateChild(nType).FirstSolidLeaf.Describe
      else  // Mark as untyped
        tmpParam.RelatedType :='Untyped';
    end;
    FreeAndNil(tmpIdentList);
  end;
end;

procedure TVInterfaceStructure.GetClassType(const ctNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ctNode);
  // Verify if it is a forward declaration
  if (lcNode.HasChildNode(nClassBody) or lcNode.HasChildNode(nClassHeritage)) then
  begin
    Assert(CurrentUnit <> nil);
    CurrentClass := CurrentUnit.AddClass(lcNode.GetParentNode(nTypeDecl).GetImmediateChild
      (nIdentifier).FirstSolidLeaf.Describe) as TDOCClass;
    CurrentClass.ClassHierarchy := 'TObject';
  end;
end;

// Interface are special classes, but no difference by now
procedure TVInterfaceStructure.GetInterfaceType(const ctNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ctNode);
  // Verify if it is a forward declaration
  if (lcNode.HasChildNode(nInterfaceBody) or lcNode.HasChildNode(nInterfaceHeritage)) then
  begin
    Assert(CurrentUnit <> nil);
    CurrentClass := CurrentUnit.AddClass(lcNode.GetParentNode(nTypeDecl).GetImmediateChild
      (nIdentifier).FirstSolidLeaf.Describe) as TDOCClass;
    CurrentClass.ClassHierarchy := 'IInterface';
  end;
end;

procedure TVInterfaceStructure.GetReturnType(const rtNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(rtNode);
  // Is a method ?
  if lcNode.HasParentNode(nFunctionReturnType) then
    CurrentMethod.RelatedType := lcNode.FirstSolidLeaf.Describe;
end;

procedure TVInterfaceStructure.GetClassHeritage(const ceNode: TObject);
var
  lcNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(ceNode);
  if lcNode.HasChildNode(nIdentList) then
  begin
    Assert(CurrentClass <> nil);
    CurrentClass.ClassHierarchy := lcNode.GetImmediateChild(nIdentList).FirstSolidLeaf.Describe;
  end;
end;

procedure TVInterfaceStructure.GetClassProperty(const cpNode: TObject);
var
  lcNode, tmpNode: TParseTreeNode;
  tmpProperty: TDOCProperty;
begin
  lcNode := TParseTreeNode(cpNode);
  // from Class declaration ?
  Assert(lcNode.HasParentNode(ObjectBodies) = True);
  Assert(CurrentClass <> nil);
  // Get Property Name
  tmpNode     := lcNode.GetImmediateChild(nIdentifier).FirstSolidLeaf;
  tmpProperty := CurrentClass.AddProperty(tmpNode.Describe) as TDOCProperty;
  tmpProperty.BodySource := GetSourceCode(lcNode);
  // Get Property Type
  tmpNode := lcNode.GetImmediateChild(nType).FirstSolidLeaf;
  tmpProperty.RelatedType := tmpNode.Describe;
  // Get Property Visibility
  GetMemberVisibility(lcNode, tmpProperty);
end;

procedure TVInterfaceStructure.GetClassField(const cfNode: TObject);
var
  lcNode, tmpNode: TParseTreeNode;
  tmpField: TDOCField;
begin
  lcNode := TParseTreeNode(cfNode);
  assert(CurrentClass <> nil);
  // Get Field Name
  tmpField := CurrentClass.AddField(lcNode.FirstSolidLeaf.Describe) as TDocField;
  tmpField.BodySource := GetSourceCode(lcNode);
  // Get Field Type
  tmpNode := lcNode.GetImmediateChild(nType).FirstSolidLeaf;
  tmpField.RelatedType := tmpNode.Describe;
  // Get Field Visibility
  GetMemberVisibility(lcNode, tmpField);
end;

procedure TVInterfaceStructure.GetUnitVar(const ivNode: TObject);
var
  lcNode: TParseTreeNode;
  tmpVar: TDOCVar;
begin
  lcNode := TParseTreeNode(ivNode);
  assert(CurrentUnit <> nil);
  tmpVar := CurrentUnit.AddVar(lcNode.FirstSolidLeaf.Describe) as TDOCVar;
  tmpVar.RelatedType := lcNode.GetImmediateChild(nType).FirstSolidLeaf.Describe;
  tmpVar.BodySource := GetSourceCode(lcNode);
  {if lcNode.HasChildNode(nVariableInit) then
  begin
    tmpNode:=lcNode.GetImmediateChild(nVariableInit).FirstSolidLeaf;
    mDebug:=mDebug+' VarInit='+tmpNode.Describe;
  end
  else mDebug:=mDebug+' VarInit=#NONE#';}
end;

procedure TVInterfaceStructure.GetUnitConst(const icNode: TObject);
var
  lcNode, tmpNode: TParseTreeNode;
  tmpConstant: TDOCConstant;
begin
  lcNode := TParseTreeNode(icNode);
  Assert(CurrentUnit <> nil);
  if lcNode.HasParentNode(nConstSection) then
  begin
    tmpConstant := CurrentUnit.AddConstant(lcNode.FirstSolidLeaf.Describe) as TDOCConstant;
    tmpConstant.BodySource := GetSourceCode(lcNode);
    if lcNode.HasChildNode(nType) then
    begin
      tmpNode := lcNode.GetImmediateChild(nType);
      if not (tmpNode.IsLeaf) then
        tmpConstant.RelatedType := tmpNode.FirstSolidLeaf.Describe
      else
        tmpConstant.RelatedType := tmpNode.Describe; //#COMPLEX#';
    end;
  end;
end;

procedure TVInterfaceStructure.GetMemberVisibility(const mvNode: TObject;
  const DocMember: TDOCMember);
var
  lcNode, tmpNode: TParseTreeNode;
begin
  lcNode := TParseTreeNode(mvNode);
  if lcNode.HasParentNode(nClassVisibility) then
  begin
    tmpNode := lcNode.GetParentNode(nClassVisibility).FirstSolidLeaf;
    DocMember.MemberVisibility := TMemberVisibility(MemberVisibilityName.IndexOf
      (TmpNode.Describe));
  end
  else if lcNode.HasParentNode(nInterfaceBody) then
    DocMember.MemberVisibility := mvPUBLIC
  else 
    DocMember.MemberVisibility := mvPUBLISHED;
end;

procedure TVInterfaceStructure.GetUnitType(const itNode: TObject);
var
  lcNode: TParseTreeNode;
  tmpType: TDOCType;
begin
  lcNode := TParseTreeNode(itNode);
  assert(CurrentUnit <> nil);
  // Take only true interface type, not Class or other
  // What about records ???
  if lcNode.HasParentNode(nTypeSection) and not lcNode.HasChildNode(nClassType) and
    not lcNode.HasChildNode(nInterfaceType) then
  begin
    tmpType := CurrentUnit.AddType(lcNode.GetImmediateChild(nIdentifier).FirstSolidLeaf.Describe)
      as TDOCType;
    tmpType.RelatedType := lcNode.GetImmediateChild(nType).FirstSolidLeaf.Describe;
    tmpType.BodySource := GetSourceCode(lcNode);
  end;
end;

end.
