{-------------------------------------------------------------------------------
  Main unit of the process side in the project
  The project, once configured through the TDOCOptions class, can Check, and
  build the output files.
  Project's state is update on each action and can be read with State property
  http://dephicodetodoc.sourceforge.net/
  Copyright TridenT 2003 - Under GNU GPL licence

  @Author    TridenT
  @Version   2003/12/01   TridenT   Initial revision
-------------------------------------------------------------------------------}

unit uDocProject;

interface

uses
  Classes, SysUtils, Structure, DocGen,uOptions;

type
  // State of the project, update on each action
  TProjectState  = (psNew,psChecked,psBuilt);

  {-----------------------------------------------------------------------------
    Main Class that manage the Parser, the Structure Builder and the Generator
     to build the Documentation from the source code in the configured format
  -----------------------------------------------------------------------------}
  TDOCProject = class
  private
    FDocOptions : TDOCOptions;  // project's Options shared with Generator.
    FDocStructure: TDOCStructure; // Holds the skeleton of the source code
    FFileList: TStringList; // List of all file that will be processed
    FState : TProjectState; // State of the project (New, Checked of Built)
    FisChecked: boolean;  // Specified if project has already been checked
    FdmUserMessages: TStrings;  // Reference for writting User messages
    FdmDebugMessages: TStrings; // Reference for writting Debug messages
    function BuildFileList: integer;
    function CheckFile(const filename: TFilename): boolean;
  public
    property DocOptions: TDOCOptions read FDocOptions write FDocOptions;
    property State : TProjectState read FState write FState;
    property isChecked: boolean read FisChecked write FisChecked;
    constructor Create;
    destructor Destroy; override;
    function Check: boolean;
    function Build: boolean;
    function GetStructure: TDOCStructure;
    procedure SetDestinationMessages(const dmUser, dmDebug: TStrings);
  end;

var
  DocProject: TDOCProject;  // The current project

implementation

uses JclFileUtils, DocGenTXT, DocGenHTML, StringsConverter, JclStrings, forms;

Resourcestring
  L_ParsingFile = 'Parsing file : %s';

{-------------------------------------------------------------------------------
  Build the project by creating the output files
  It verify if the project is ready to be build, and then call the Generator
  by the execute method.
  Additionnaly, the project's State is set to psBuilt if successfull
  @return   TRUE if succesfull, FALSE otherwise
-------------------------------------------------------------------------------}
function TDOCProject.Build: boolean;
var
  DocGenerator: TDOCGenerator;
begin
  if DocProject.isChecked then
  begin
    // Create Generator
    DocGenerator := TDocGenHTML.Create;
    DocGenerator.DocOptions:= FDocOptions;
    try
      // Build
      DocGenerator.Execute(DocProject.FDocStructure);
    finally
      // Free
      FreeAndNil(DocGenerator);
    end;
    Result := True;
    FState := psBuilt;
  end
  else
    Result := False;
end;

// NOT MORE USED
function PasFilter(const Attr: integer; const FileInfo: TSearchRec): boolean;
begin
  Result := (AnsiUpperCase(ExtractFileExt(FileInfo.Name)) = '.PAS');
end;

{-------------------------------------------------------------------------------
  Build a file list filtered with PAS extension
  Start is the input directory, a,s eventually the sub-directoryes,
  it then add all mathing files in the FileList.
  @return   Number of files found if succesfull, -1 otherwise
-------------------------------------------------------------------------------}
function TDOCProject.BuildFileList: integer;
var
  flOptions: TFileListOptions;
begin
  if FDocOptions.RecurseDirectory then flOptions := [flRecursive, flFullNames]
  else 
    flOptions := [flFullNames];
  // $80 must be added because if the file's archive attribute is not set,
  // then FindFirst return [FindoInfo.Attr = 128]  ...
  if AdvBuildFileList(FDocOptions.RootDirectory + '*.pas',
    $80 + faReadOnly + faHidden + faSysFile + faArchive
    , FFileList,{amSubsetOf} amAny, flOptions, '', nil) then
  begin
    Result := FFileList.Count;
  end
  else 
    Result := -1;
end;

{-------------------------------------------------------------------------------
  Check if Project can be build
  This is a must before building the project. It checks and parse all files in
  the FileList.
  If successfull, the Projects's state is set to psChecked
  @return     TRUE if succesfull, FALSE otherwise
-------------------------------------------------------------------------------}
function TDOCProject.Check: boolean;
var
  IndexFile: integer;
  ProcessedFiles: integer;
begin
  // Clear all units
  FDocStructure.Clear;
  FFileList.Clear;

  ProcessedFiles     := 0;
  FDocStructure.Name := FDocOptions.Name;
  // Create Filelist
  if (BuildFileList<>-1) then
  begin
    // Configure Parser and Builder
    // parse and build structure
    for IndexFile := 0 to FFileList.Count - 1 do
    begin
      if CheckFile(FFileList.Strings[IndexFile]) then
        ProcessedFiles := ProcessedFiles + 1;
    end;
    Result := (ProcessedFiles = FFileList.Count);
    if FDocOptions.AlphaSort then FDocStructure.SortContent;
    FState := psChecked;
  end
  else 
    Result := False;
  FisChecked := Result;
end;

{-------------------------------------------------------------------------------
  Check and parse a delphi source code file
  For the moment, this function always return a successfull value
  @param    filename  File to Check
  @return   TRUE if succesfull, FALSE otherwise
-------------------------------------------------------------------------------}
function TDOCProject.CheckFile(const filename: TFilename): boolean;
var
  slInput: TStrings;
  fcConvert: TStringsConverter;
begin
  FdmDebugMessages.Add(Format(L_ParsingFile, [filename]));

  slInput := TStringList.Create;
  slInput.LoadFromFile(filename);

  fcConvert := TStringsConverter.Create;
  fcConvert.InputStrings := slInput;
  fcConvert.OutputStrings := FdmUserMessages;
  fcConvert.MessageStrings := FdmDebugMessages;
  fcConvert.Convert;
  fcConvert.Clear;

  FreeAndNil(slInput);
  FreeAndNil(fcConvert);
  Result := True;
end;

{-------------------------------------------------------------------------------
  Creates and initializes a TDOCProject
  The project's state is set to psNew after the creation
-------------------------------------------------------------------------------}
constructor TDOCProject.Create;
begin
  inherited;
  FDocOptions      := TDOCOptions.Create(Application);
  FDocStructure := TDOCStructure.Create('', nil);
  FFileList     := TStringList.Create;
  FState := psNew;
end;

{-------------------------------------------------------------------------------
  Destroy a TDOCProject reference
  It frees all lists its own
-------------------------------------------------------------------------------}
destructor TDOCProject.Destroy;
begin
  FreeAndNil(FFileList);
  FreeAndNil(FDocStructure);
  FreeAndNil(FDocOptions);
  inherited;
end;

{-------------------------------------------------------------------------------
  Return the Structure of a project
  @return   Structure reference
-------------------------------------------------------------------------------}
function TDOCProject.GetStructure: TDOCStructure;
begin
  Result := FDocStructure;
end;

{-------------------------------------------------------------------------------
  Configure the User output and Debug output to inform about the status
  The TDOCProject class is very separated from the external user interface.
  But in order to inform about the state of the project, it needs to know where
  to display informations. This could be a visual control, or a file or whatever
  containing a TStrings reference to work.

  @param    dmUser  Where User messages will be written
  @param    dmDebug Where Debug messages will be written
-------------------------------------------------------------------------------}
procedure TDOCProject.SetDestinationMessages(const dmUser, dmDebug: TStrings);
begin
  FdmUserMessages  := dmUser;
  FdmDebugMessages := dmDebug;
end;



end.
